#include <cnoid/SimpleController>

#include <cnoid/Body>
#include <cnoid/BodyLoader>
#include <cnoid/VRMLBodyLoader>
#include <cnoid/EigenTypes>
#include <cnoid/JointPath>
#include <cnoid/Jacobian>
#include <cnoid/EigenUtil>
#include <cnoid/Sensor>
#include <cnoid/Link>

#include <vector>

#include <string>
#include <iostream>
#include <iomanip>
#include <cstring>

using namespace cnoid;
using namespace std;

//P制御ゲイン
const double pgain[] = { 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0, 8000.0 };
//D制御ゲイン
const double dgain[] = { 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, };


//SimpleControllerクラスのオーバーロード
class Controller: public SimpleController {

	//各種値の初期化//
	cnoid::BodyPtr robot;		//Bodyクラスへのスマートポインタ(入出力に使う)
	double dt;						//制御系のサンプリングタイム

	std::vector<double> qref;
	std::vector<double> qold;




public:

	virtual bool initialize(SimpleControllerIO *io) override {
		robot = io->body();
		dt = io->timeStep();

		for (int i = 0; i < robot->numJoints(); ++i) {
			Link *joint = robot->joint(i);
			joint->setActuationMode(Link::JOINT_TORQUE);
			io->enableIO(joint);						
			qref.push_back(joint->q());
		}

		qold = qref;

		//桁表示の固定
		std::cout << std::fixed;

		return true;
	}

	virtual bool control() override {
		//慣性行列の初期化
		Eigen::MatrixXd MM = Eigen::MatrixXd::Zero(robot->numJoints() + 6, robot->numJoints() + 6);
		//慣性行列の計算
		calcMassMatrix(robot, MM);

		cout << "慣性行列 M =" << endl;
		cout << std::setprecision(2) << MM << endl;

		for (int i = 0; i < robot->numJoints(); ++i) {
			Link *joint = robot->joint(i);
			double q = joint->q();
			double dq = (q - qold[i]) / dt;
			double u = (qref[i] - q) * pgain[i] + (0.0 - dq) * dgain[i];
			qold[i] = q;
			joint->u() = u;
			cout << u << endl;
		}
		return true;
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(Controller)

